@@ -2,24 +2,29 @@ package POE::Wheel::SocketFactory;
 
 use strict;
 
-use vars qw($VERSION);
-$VERSION = '1.284'; # NOTE - Should be #.### (three decimal places)
+use vars qw($VERSION @ISA);
+$VERSION = '1.311'; # NOTE - Should be #.### (three decimal places)
 
 use Carp qw( carp croak );
 use Symbol qw( gensym );
 
 use Fcntl qw(F_GETFL F_SETFL O_NONBLOCK);
-use Errno qw(EWOULDBLOCK EADDRNOTAVAIL EINPROGRESS EADDRINUSE);
+use Errno qw(
+  EWOULDBLOCK EADDRNOTAVAIL EINPROGRESS EADDRINUSE ECONNABORTED
+  ESPIPE
+);
+
 use Socket qw(
   AF_INET SOCK_STREAM SOL_SOCKET AF_UNIX PF_UNIX 
   PF_INET SOCK_DGRAM SO_ERROR unpack_sockaddr_in 
   unpack_sockaddr_un PF_UNSPEC SO_REUSEADDR INADDR_ANY 
   pack_sockaddr_in pack_sockaddr_un inet_aton SOMAXCONN
 );
+
 use IO::Handle ();
 use FileHandle ();
 use POE qw( Wheel );
-use base qw(POE::Wheel);
+push @ISA, qw(POE::Wheel);
 
 sub CRIMSON_SCOPE_HACK ($) { 0 }
 sub DEBUG () { 0 }
@@ -38,25 +43,38 @@ sub MY_SOCKET_TYPE     () { 10 }
 sub MY_STATE_ERROR     () { 11 }
 sub MY_SOCKET_SELECTED () { 12 }
 
-# Fletch has subclassed SSLSocketFactory from SocketFactory.  He's
-# added new members after MY_SOCKET_SELECTED.  Be sure, if you extend
-# this, to extend add stuff BEFORE MY_SOCKET_SELECTED or let Fletch
-# know you've broken his module.
+# Fletch has subclassed SSLSocketFactory from SocketFactory.
+# He's added new members after MY_SOCKET_SELECTED.  Be sure, if you
+# extend this, to extend add stuff BEFORE MY_SOCKET_SELECTED or let
+# Fletch know you've broken his module.
 
-# Provide dummy constants for systems that don't have them.
-# Test and provide for each constant separately, per suggestion in
-# rt.cpan.org 27250.
+# If IPv6 support can't be loaded, then provide dummies so the code at
+# least compiles.  Suggested in rt.cpan.org 27250.
 BEGIN {
-  eval { require Socket6 };
+  # under perl-5.6.2 the warning "leaks" from the eval, while newer versions don't...
+  # it's due to Exporter.pm behaving differently, so we have to shut it up
+  no warnings 'redefine';
+  local *Carp::carp = sub { die @_ };
+
+  # Socket::GetAddrInfo provides getaddrinfo where earlier Perls' Socket don't.
+  eval { Socket->import( qw(getaddrinfo getnameinfo) ) };
   if ($@) {
-    *Socket6::AF_INET6 = sub () { ~0 };
-    *Socket6::PF_INET6 = sub () { ~0 };
+    # :newapi is legacy, but we include it to be sure in case the user has an old version of GAI
+    eval { require Socket::GetAddrInfo; Socket::GetAddrInfo->import( qw(:newapi getaddrinfo getnameinfo) ) };
+    if ($@) {
+      *getaddrinfo = sub { Carp::confess("Unable to use IPv6: Socket::GetAddrInfo not available") };
+      *getnameinfo = sub { Carp::confess("Unable to use IPv6: Socket::GetAddrInfo not available") };
+    }
   }
-  else {
-    eval { my $x = &Socket6::AF_INET6 };
-    *Socket6::AF_INET6 = sub () { ~0 } if $@;
-    eval { my $x = &Socket6::PF_INET6 };
-    *Socket6::PF_INET6 = sub () { ~0 } if $@;
+
+  # Socket6 provides AF_INET6 and PF_INET6 where earlier Perls' Socket don't.
+  eval { Socket->import( qw(AF_INET6 PF_INET6) ) };
+  if ($@) {
+    eval { require Socket6; Socket6->import( qw(AF_INET6 PF_INET6) ) };
+    if ($@) {
+      *AF_INET6 = sub { Carp::confess("Unable to use IPv6: Socket6 not available") };
+      *PF_INET6 = sub { Carp::confess("Unable to use IPv6: Socket6 not available") };
+    }
   }
 }
 
@@ -73,8 +91,8 @@ sub DOM_INET6 () { 'inet6' }  # INET v6 domain socket
 my %map_family_to_domain = (
   AF_UNIX,  DOM_UNIX,  PF_UNIX,  DOM_UNIX,
   AF_INET,  DOM_INET,  PF_INET,  DOM_INET,
-  &Socket6::AF_INET6, DOM_INET6,
-  &Socket6::PF_INET6, DOM_INET6,
+  AF_INET6, DOM_INET6,
+  PF_INET6, DOM_INET6,
 );
 
 sub SVROP_LISTENS () { 'listens' }  # connect/listen sockets
@@ -171,7 +189,8 @@ sub _define_accept_state {
         }
         elsif ( $domain eq DOM_INET6 ) {
           $peer = getpeername($new_socket);
-          ($peer_port, $peer_addr) = Socket6::unpack_sockaddr_in6($peer);
+          ((my $error), $peer_port, $peer_addr) =  getnameinfo($peer);
+          warn $error if $error;
         }
         else {
           die "sanity failure: socket domain == $domain";
@@ -182,7 +201,8 @@ sub _define_accept_state {
           $unique_id
         );
       }
-      elsif ($! != EWOULDBLOCK) {
+      elsif ($! != EWOULDBLOCK and $! != ECONNABORTED and $! != ESPIPE) {
+        # OSX reports ESPIPE, which isn't documented anywhere.
         $$event_failure && $k->call(
           $me, $$event_failure,
           'accept', ($!+0), $!, $unique_id
@@ -289,10 +309,9 @@ sub _define_connect_state {
       # INET6 socket stacks tend not to.
       elsif ($domain eq DOM_INET6) {
         if (defined $peer) {
-          eval {
-            ($peer_port, $peer_addr) = Socket6::unpack_sockaddr_in6($peer);
-          };
-          if (length $@) {
+          ((my $error), $peer_port, $peer_addr) = getnameinfo($peer);
+          if ($error) {
+            warn $error;
             $peer_port = $peer_addr = undef;
           }
         }
@@ -311,9 +330,9 @@ sub _define_connect_state {
     }
   );
 
-  # Cygwin expects an error state registered to expedite.  This code
-  # is nearly identical the stuff above.
-  if ($^O eq "cygwin") {
+  # Cygwin and Windows expect an error state registered to expedite.
+  # This code is nearly identical the stuff above.
+  if ($^O eq "cygwin" or $^O eq "MSWin32") {
     $poe_kernel->state(
       $self->[MY_STATE_ERROR] = (
         ref($self) .  "($unique_id) -> connect error"
@@ -405,7 +424,7 @@ sub event {
       $self->[MY_SOCKET_HANDLE],
       $self->[MY_STATE_CONNECT]
     );
-    if ($^O eq "cygwin") {
+    if ($^O eq "cygwin" or $^O eq "MSWin32") {
       $poe_kernel->select_expedite(
         $self->[MY_SOCKET_HANDLE],
         $self->[MY_STATE_ERROR]
@@ -421,7 +440,10 @@ sub event {
 
 sub getsockname {
   my $self = shift;
-  return undef unless defined $self->[MY_SOCKET_HANDLE];
+  return undef unless (
+    defined $self->[MY_SOCKET_HANDLE] and
+    fileno($self->[MY_SOCKET_HANDLE])
+  );
   return getsockname($self->[MY_SOCKET_HANDLE]);
 }
 
@@ -628,58 +650,7 @@ sub new {
 
   # Don't block on socket operations, because the socket will be
   # driven by a select loop.
-
-  # RCC 2002-12-19: Replace the complex blocking checks and methods
-  # with IO::Handle's blocking(0) method.  This is theoretically more
-  # portable and less maintenance than rolling our own.  If things
-  # work out, we'll remove the commented out code.
-
-  # RCC 2003-01-20: Unfortunately, blocking() isn't available in perl
-  # 5.005_03, and people still use that.  We'll use blocking() for
-  # Perl 5.8.0 and beyond, since that's the first version of
-  # ActivePerl that has a problem.
-
-  if ($] >= 5.008) {
-    $socket_handle->blocking(0);
-  }
-  else {
-    # Do it the Win32 way.  XXX This is incomplete.
-    if ($^O eq 'MSWin32') {
-      my $set_it = "1";
-
-      # 126 is FIONBIO (some docs say 0x7F << 16)
-      ioctl(
-        $socket_handle,
-        0x80000000 | (4 << 16) | (ord('f') << 8) | 126,
-        \$set_it
-      ) or do {
-        $poe_kernel->yield(
-          $event_failure,
-          'ioctl', $!+0, $!, $self->[MY_UNIQUE_ID]
-        );
-        return $self;
-      };
-    }
-
-    # Do it the way everyone else does.
-    else {
-      my $flags = fcntl($socket_handle, F_GETFL, 0) or do {
-        $poe_kernel->yield(
-          $event_failure,
-          'fcntl', $!+0, $!, $self->[MY_UNIQUE_ID]
-        );
-        return $self;
-      };
-
-      $flags = fcntl($socket_handle, F_SETFL, $flags | O_NONBLOCK) or do {
-        $poe_kernel->yield(
-          $event_failure,
-          'fcntl', $!+0, $!, $self->[MY_UNIQUE_ID]
-        );
-        return $self;
-      };
-    }
-  }
+  $socket_handle->blocking(0);
 
   # Make the socket reusable, if requested.
   if (
@@ -776,7 +747,7 @@ sub new {
       $bind_address = (
         (defined $params{BindAddress})
         ? $params{BindAddress}
-        : Socket6::in6addr_any()
+        : "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"  # XXX - Only Socket6 has?
       );
 
       # Set the bind port, or default to 0 (any) if none specified.
@@ -798,12 +769,16 @@ sub new {
       BEGIN { eval { require bytes } and bytes->import; }
 
       # Resolve the bind address.
-      my @info = Socket6::getaddrinfo(
-        $bind_address, $bind_port,
-        $self->[MY_SOCKET_DOMAIN], $self->[MY_SOCKET_TYPE],
+      my ($error, @addresses) = getaddrinfo(
+        $bind_address, $bind_port, {
+          family   => $self->[MY_SOCKET_DOMAIN],
+          socktype => $self->[MY_SOCKET_TYPE],
+        }
       );
 
-      if (@info < 5) {  # unless defined $bind_address
+      unless (@addresses) {
+        warn $error if $error;
+
         $! = EADDRNOTAVAIL;
         $poe_kernel->yield(
           $event_failure,
@@ -812,7 +787,7 @@ sub new {
         return $self;
       }
 
-      $bind_address = $info[3];
+      $bind_address = $addresses[0]->{addr};
     }
   }
 
@@ -900,16 +875,19 @@ sub new {
         $error_tag = "inet_aton";
       }
       elsif ($abstract_domain eq DOM_INET6) {
-        my @info = Socket6::getaddrinfo(
-          $params{RemoteAddress}, $remote_port,
-          $self->[MY_SOCKET_DOMAIN], $self->[MY_SOCKET_TYPE],
+        my ($error, @addresses) = getaddrinfo(
+          $params{RemoteAddress}, $remote_port, {
+            family   => $self->[MY_SOCKET_DOMAIN],
+            socktype => $self->[MY_SOCKET_TYPE],
+          },
         );
 
-        if (@info < 5) {
+        unless (@addresses) {
+          warn $error if $error;
           $connect_address = undef;
         }
         else {
-          $connect_address = $info[3];
+          $connect_address = $addresses[0]->{addr};
         }
 
         $error_tag = "getaddrinfo";
@@ -1222,7 +1200,7 @@ work where establishing connections normally would block.
 
 =head2 new
 
-new() creats a new POE::Wheel::SocketFactory object.  For sockets
+new() creates a new POE::Wheel::SocketFactory object.  For sockets
 which listen() for and accept() connections, the wheel will generate
 new sockets for each accepted client.  Socket factories for one-shot
 sockets, such as UDP peers or clients established by connect() only
@@ -1253,11 +1231,14 @@ POE::Wheel::SocketFactory contains a table of supported domains and
 the instructions needed to create them.  Please send patches to
 support additional domains, as needed.
 
-Note: C<AF_INET6> and C<PF_INET6> are supplied by the L<Socket6>
-module, which is available on the CPAN.  You must have Socket6 loaded
-before SocketFactory can create IPv6 sockets.
+Note: C<AF_INET6> and C<PF_INET6> are supplied by the L<Socket>
+module included in Perl 5.8.0 or later.  Perl versions before 5.8.0
+should not attempt to use IPv6 until someone contributes a workaround.
 
-TODO - Example.
+IPv6 support requires a 21st century Socket module and the presence of
+Socket::GetAddrInfo to resolve host names to IPv6 addresses.
+
+Z<TODO - Example.>
 
 =head4 SocketType
 
@@ -1265,7 +1246,7 @@ C<SocketType> supplies the socket() call with a particular socket
 type, which may be C<SOCK_STREAM> or C<SOCK_DGRAM>.  C<SOCK_STREAM> is
 the default if C<SocketType> is not supplied.
 
-TODO - Example.
+Z<TODO - Example.>
 
 =head4 SocketProtocol
 
@@ -1276,7 +1257,7 @@ domain sockets.
 The protocol defaults to "tcp" for INET domain sockets.  There is no
 default for other socket domains.
 
-TODO - Example.
+Z<TODO - Example.>
 
 =head3 Setting Socket Options
 
@@ -1325,7 +1306,7 @@ contain a path describing the socket's filename.  This is required for
 server sockets and datagram client sockets.  C<BindAddress> has no
 default value for UNIX sockets.
 
-TODO - Example.
+Z<TODO - Example.>
 
 =head4 BindPort
 
@@ -1337,7 +1318,7 @@ choose an indeterminate unallocated port.
 C<BindPort> may be a port number or a name that can be looked up in
 the system's services (or equivalent) database.
 
-TODO - Example.
+Z<TODO - Example.>
 
 =head3 Connectionless Sockets
 
@@ -1346,11 +1327,11 @@ needing to listen() for connections or connect() to remote addresses.
 
 This class of sockets is complete after the bind() call.
 
-TODO - Example.
+Z<TODO - Example.>
 
 =head3 Connecting the Socket to a Remote Endpoint
 
-A socket may either listen for connections to arrive, intitate
+A socket may either listen for connections to arrive, initiate
 connections to a remote endpoint, or be connectionless (such as in the
 case of UDP sockets).
 
@@ -1377,7 +1358,7 @@ Internet address, or a UNIX socket path.  It will be packed, with or
 without an accompanying C<RemotePort>, as necessary for the socket
 domain.
 
-TODO - Example.
+Z<TODO - Example.>
 
 =head4 RemotePort
 
@@ -1388,7 +1369,7 @@ contain both an address and a port.
 The remote port may be numeric, or it may be a symbolic name found in
 /etc/services or the equivalent for your operating system.
 
-TODO - Example.
+Z<TODO - Example.>
 
 =head3 Listening for Connections
 
@@ -1408,7 +1389,7 @@ than C<SOMAXCONN> will be clipped to C<SOMAXCONN>.  Excessively large
 C<ListenQueue> values are not necessarily portable, and may cause
 errors in some rare cases.
 
-TODO - Example.
+Z<TODO - Example.>
 
 =head3 Emitting Events
 
@@ -1435,20 +1416,20 @@ client.
 =head4 FailureEvent
 
 C<FailureEvent> names the event POE::Wheel::SocketFactory will emit
-whenever something goes wrong.  It ususally represents some kind of
+whenever something goes wrong.  It usually represents some kind of
 built-in function call error.  See L</PUBLIC EVENTS> for details, as
 some errors are handled internally by this wheel.
 
 =head2 event
 
 event() allows a session to change the events emitted by a wheel
-without destrying and re-creating the wheel.  It accepts one or more
+without destroying and re-creating the wheel.  It accepts one or more
 of the events listed in L</PUBLIC EVENTS>.  Undefined event names
 disable those events.
 
 event() is described in more depth in L<POE::Wheel>.
 
-TODO - Example.
+Z<TODO - Example.>
 
 =head2 getsockname
 
@@ -1462,7 +1443,7 @@ to which POE::Wheel::SocketFactory has bound its listening socket.
 Test applications may use getsockname() to find the server socket
 after POE::Wheel::SocketFactory has bound to INADDR_ANY port 0.
 
-TODO - Example.
+Z<TODO - Example.>
 
 =head2 ID
 
@@ -1470,7 +1451,7 @@ ID() returns the wheel's unique ID.  The ID will also be included in
 every event the wheel generates.  Applications can match events back
 to the objects that generated them.
 
-TODO - Example.
+Z<TODO - Example.>
 
 =head2 pause_accept
 
@@ -1487,7 +1468,7 @@ pause_accept() and resume_accept() is quicker and more reliable than
 dynamically destroying and re-creating a POE::Wheel::SocketFactory
 object.
 
-TODO - Example.
+Z<TODO - Example.>
 
 =head2 resume_accept
 
@@ -1518,7 +1499,9 @@ listening or connecting.  See below for the differences.
 
 For INET sockets, C<$_[ARG1]> and C<$_[ARG2]> hold the socket's remote
 address and port, respectively.  The address is packed; see
-L<Socket/inet_nota()> if a human-readable version is needed.
+L<Socket/inet_ntoa> if a human-readable IPv4 address is needed.
+L<Socket::GetAddrInfo/getnameinfo> provides numeric addresses for IPv4
+and IPv6 addresses.
 
   sub handle_new_client {
     my $accepted_socket = $_[ARG0];
@@ -1557,7 +1540,8 @@ holds a Perl built-in function name.
 
 C<$_[ARG1]> and C<$_[ARG2]> describe how the operation failed.  They
 contain the numeric and stringified versions of C<$!>, respectively.
-An application cannot merely check C<$!> because
+An application cannot merely check the global C<$!> variable since it
+may change during event dispatch.
 
 Finally, C<$_[ARG3]> contains the ID for the POE::Wheel::SocketFactory
 instance that generated the event.  See L</ID> and L<POE::Wheel/ID>
@@ -1576,9 +1560,12 @@ A sample FailureEvent handler:
 L<POE::Wheel> describes the basic operations of all wheels in more
 depth.  You need to know this.
 
-L<Socket6> is required for IPv6 work.  POE::Wheel::SocketFactory will
-load it automatically if it's installed, but applications will need to
-use it themselves to get access to AF_INET6.
+L<Socket::GetAddrInfo> is required for IPv6 work.
+POE::Wheel::SocketFactory will load it automatically if it's
+installed.  SocketDomain => AF_INET6 is required to trigger IPv6
+behaviors.  AF_INET6 is exported by the Socket module on all but the
+oldest versions of Perl 5.  If your Socket doesn't provide AF_INET6,
+try installing Socket6 instead.
 
 The SEE ALSO section in L<POE> contains a table of contents covering
 the entire POE distribution.
